/////////////////////////////////////////////////////////////////////////////
// Copyright (c) 1996, California Institute of Technology.  
// ALL RIGHTS RESERVED.   U.S. Government Sponsorship acknowledged. 
//
// Please read the full copyright notice in the file COPYRIGH 
// in this directory.
//
// Author: Todd Karakashian, NASA/Jet Propulsion Laboratory
//         Todd.K.Karakashian@jpl.nasa.gov
//
// $RCSfile: dodsutil.cc,v $ - Miscellaneous routines for DODS HDF server
//
// $Log: dodsutil.cc,v $
// Revision 1.4.8.1  2000/03/09 00:24:59  jimg
// Replaced int and uint32 with string::size_type
//
// Revision 1.4  1999/05/06 03:23:35  jimg
// Merged changes from no-gnu branch
//
// Revision 1.3.6.1  1999/05/06 00:27:23  jimg
// Jakes String --> string changes
//
// Revision 1.2  1998/02/05 20:14:30  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.1  1997/03/10 22:55:01  jimg
// New files for the 2.12 compatible HDF server
//
// Revision 1.3  1996/11/20  22:28:43  todd
// Modified to support UInt32 type.
//
// Revision 1.2  1996/10/07 21:15:17  todd
// Changes escape character to % from _.
//
// Revision 1.1  1996/09/24 22:38:16  todd
// Initial revision
//
//
/////////////////////////////////////////////////////////////////////////////

#include <strstream.h>
#include <iomanip>
#include <string>
#include <Regex.h>
#include "HDFStructure.h"
#include "HDFArray.h"
#include "dodsutil.h"
#include "dhdferr.h"
#include "hdf-dods.h"

int regindex(const string &s, const Regex &r, int index = 0) {
  int matchlen;
  return r.search(s.c_str(), s.size(), matchlen, index);
}

string hexstring(unsigned char val) {
    static char buf[hdfdods::MAXSTR];

    ostrstream(buf,hdfdods::MAXSTR) << hex << setw(2) << setfill('0') << 
	(unsigned int)val << ends;

    return (string)buf;
}

char unhexstring(string s) {
    int val;

    istrstream(s.c_str(),hdfdods::MAXSTR) >> hex >> val;

    return (char)val;
}

string octstring(unsigned char val) {
    static char buf[hdfdods::MAXSTR];

    ostrstream(buf,hdfdods::MAXSTR) << oct << setw(3) << setfill('0') << 
	(unsigned int)val << ends;

    return (string)buf;
}

char unoctstring(string s) {
    int val;

    istrstream(s.c_str(),hdfdods::MAXSTR) >> oct >> val;

    return (char)val;
}


// replace characters that are not allowed in DODS identifiers
string id2dods(string s) {
    static Regex badregx("[^0-9a-zA-Z_%]");
    static const string ESC = "%";

    int index=0;
    while ( (index = regindex(s, badregx, index)) >= 0)
	s.replace(index, 1, ESC + hexstring(s[index]));
    if (isdigit(s[0]))
	s.insert(0, '_');
    return s;
}

string dods2id(string s) {
    static Regex escregx("%[0-7][0-9a-fA-F]");

    int index=0;
    while ( (index = regindex(s, escregx, index)) != -1)
	s.replace(index, 3, unhexstring(s.substr(index+1,2)));

    return s;
}

// Escape non-printable characters and quotes from an HDF attribute
string escattr(string s) {
    static Regex nonprintable("[^ !-~]");
    const string ESC = "\\";
    const char QUOTE = '\"';
    const string ESCQUOTE = ESC + QUOTE;

    // escape non-printing characters with octal escape
    string::size_type index = 0;
    while ( (index = regindex(s, nonprintable, index)) >= 0)
	s.replace(index, 1, ESC + octstring(s[index]));

    // escape " with backslash
    index = 0;
    while ( (index = s.find(QUOTE, index)) != s.npos) {
	s.replace(index,1, ESCQUOTE);
	index += ESCQUOTE.length();
    }
	   
    return s;
}

// Un-escape special characters, quotes and backslashes from an HDF attribute.
// Note: A regex to match one \ must be defined as
//          Regex foo = "\\\\";
//       because both C++ strings and libg++ regex's also employ \ as
//       an escape character!
string unescattr(string s) {
    static Regex escregx("\\\\[01][0-7][0-7]");  // matches 4 characters
    static Regex escquoteregex("[^\\\\]\\\\\"");  // matches 3 characters
    static Regex escescregex("\\\\\\\\");      // matches 2 characters
    const string ESC = "\\";
    const char QUOTE = '\"';
    const string ESCQUOTE = ESC + QUOTE;

    // unescape any octal-escaped ASCII characters
    int index = 0;
    while ( (index = regindex(s,escregx,index)) >= 0) {
	s.replace(index,4, unoctstring(s.substr(index+1,3)));
	index++;
    }

    // unescape any escaped quotes
    index = 0;
    while ( (index = regindex(s,escquoteregex,index)) >= 0) {
	s.replace(index+1,2, QUOTE);
	index++;
    }

    // unescape any escaped backslashes
    index = 0;
    while ( (index = regindex(s,escescregex,index)) >= 0) {
	s.replace(index,2, ESC);
	index++;
    }
    
    return s;
}

HDFStructure *CastBaseTypeToStructure(BaseType *p) {
    if (p->type_name() != "Structure")
	THROW(dhdferr_consist);
    return (HDFStructure *)p;
}

HDFArray *CastBaseTypeToArray(BaseType *p) {
    if (p->type_name() != "Array")
	THROW(dhdferr_consist);
    return (HDFArray *)p;
}

int *CastBaseTypeToInt(BaseType *p) {
    if (p->type_name() != "Int32")
	THROW(dhdferr_consist);
    return (int *)p;
}

double *CastBaseTypeToDouble(BaseType *p) {
    if (p->type_name() != "Float64")
	THROW(dhdferr_consist);
    return (double *)p;
}

